/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 * Paul Mahar
 *
 */
package org.enhydra.kelp.common.importer;

// ToolBox imports
import org.enhydra.tool.common.PathHandle;
import org.enhydra.tool.common.SwingUtil;

// Kelp imports
import org.enhydra.kelp.common.PathUtil;
import org.enhydra.kelp.common.ValidationException;
import org.enhydra.kelp.common.ValidationUtil;

// Standard imports
import javax.swing.*;
import javax.swing.border.*;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.*;
import java.io.File;
import java.util.ResourceBundle;

/**
 * Panel for entering the default options: project name, package
 * and destination.
 *
 * @author Paul Mahar
 */
public class PathPanel extends ImporterPanel {
    static ResourceBundle                 res =
        ResourceBundle.getBundle("org.enhydra.kelp.common.Res");    // nores
    transient private JPanel              panelSource;
    transient private JTextField          textSource;
    transient private JButton             buttonSource;
    transient private JPanel              panelResource;
    transient private JTextField          textResource;
    transient private JButton             buttonResource;

    //
    transient private JPanel              panelFiller;
    transient private GridBagLayout       layoutMain;
    transient private GridBagLayout       layoutSource;
    transient private GridBagLayout       layoutResource;
    transient private LocalButtonListener buttonListener;
    transient private TitledBorder        borderSource;
    transient private TitledBorder        borderResource;

    /**
     * Create default option panel 1 for entering the
     * following default options: project name, package and
     * destination.
     */
    public PathPanel() {
        try {
            jbInit();
            pmInit();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    /**
     * Read project name, package and destination values from the
     * option set into the swing controls.
     *
     * @exception GeneratorException
     * Thrown if unable to update Swing controls with option set values.
     */
    public void readOptions() {

        // wait till activated to to any real reading
    }

    /**
     * Write project name, package and destination values from the
     * swing controls into the option set.
     *
     * @exception GeneratorException
     * Thrown if unable to update option set from Swing control values.
     */
    public void writeOptions() throws ValidationException {
        String value = new String();

        if (getPaths() == null) {
            throw new ValidationException(res.getString("Import_paths_not_set"));
        }

        // source
        value = textSource.getText().trim();
        getPaths().setSourcePath(value);

        // resource
        value = textResource.getText().trim();
        getPaths().setResourcePath(value);
    }

    /**
     * Validate project name, package and destination values
     * in the swing controls.
     *
     * @exception ValidationException
     * Thrown if swing control values are not valid for the
     * current option set.
     */
    public void validateOptions() throws ValidationException {
        String value = new String();

        // package
        value = textSource.getText().trim();
        if (!ValidationUtil.isDirectory(value)) {
            throw new ValidationException(res.getString("SourcePathInvalid")
                                          + res.getString("DirectoryNotFound"));
        }

        // destination
        value = textResource.getText().trim();
        if (!ValidationUtil.isDirectory(value)) {
            throw new ValidationException(res.getString("ResourcePathInvalid")
                                          + res.getString("DirectoryNotFound"));
        }
    }

    /**
     * Get the title to use on the current page.
     *
     * @return
     * A string to place at the top of a CodeGen wizard panel.
     */
    public String getPageTitle() {
        return res.getString("Path_details");
    }

    /**
     * Get the instructions for entering option values for the
     * current page.
     *
     * @return
     * A string to place below the page title.
     */
    public String getInstructions() {
        StringBuffer buf = new StringBuffer();

        buf.append(res.getString("PathInstruct1"));
        buf.append(res.getString("PathInstruct2"));
        buf.append(res.getString("PathInstruct3"));
        buf.append(res.getString("PathInstruct4"));
        buf.append(res.getString("PathInstruct5"));
        buf.append(res.getString("PathInstruct6"));
        buf.append(res.getString("PathInstruct7"));
        buf.append(res.getString("PathInstruct8"));
        buf.append(res.getString("PathInstruct9"));
        return buf.toString();
    }

    public void activated() {
        initPaths();
    }

    private void initPaths() {
        String path = null;

        // source path
        path = getPaths().getSourcePath();
        if (path == null) {
            textSource.setText(getPaths().getDefaultSourcePath());
        } else {
            textSource.setText(path);
        }

        // resource path
        path = getPaths().getResourcePath();
        if (path == null) {
            textResource.setText(getPaths().getDefaultResourcePath());
        } else {
            textResource.setText(path);
        }
    }

    // /
    // /

    /**
     * Method declaration
     *
     *
     * @exception Exception
     *
     * @see
     */
    private void jbInit() throws Exception {
        textSource =
            (JTextField) Beans.instantiate(getClass().getClassLoader(),
                                           JTextField.class.getName());
        panelSource = (JPanel) Beans.instantiate(getClass().getClassLoader(),
                                                 JPanel.class.getName());
        panelResource =
            (JPanel) Beans.instantiate(getClass().getClassLoader(),
                                       JPanel.class.getName());
        layoutSource =
            (GridBagLayout) Beans.instantiate(getClass().getClassLoader(),
                                              GridBagLayout.class.getName());
        layoutResource =
            (GridBagLayout) Beans.instantiate(getClass().getClassLoader(),
                                              GridBagLayout.class.getName());
        buttonSource =
            (JButton) Beans.instantiate(getClass().getClassLoader(),
                                        JButton.class.getName());
        buttonResource =
            (JButton) Beans.instantiate(getClass().getClassLoader(),
                                        JButton.class.getName());
        textResource =
            (JTextField) Beans.instantiate(getClass().getClassLoader(),
                                           JTextField.class.getName());
        layoutMain =
            (GridBagLayout) Beans.instantiate(getClass().getClassLoader(),
                                              GridBagLayout.class.getName());
        panelFiller = (JPanel) Beans.instantiate(getClass().getClassLoader(),
                                                 JPanel.class.getName());
        buttonSource.setText("..."); // nores
        buttonResource.setText("...");  // nores
        panelSource.setLayout(layoutSource);
        panelSource.add(textSource,
                        new GridBagConstraints(0, 0, 1, 1, 0.1, 0.1,
                                               GridBagConstraints.WEST,
                                               GridBagConstraints.HORIZONTAL,
                                               new Insets(5, 5, 5, 5), 0, 0));
        panelSource.add(buttonSource,
                        new GridBagConstraints(1, 0, 1, 2, 0.0, 0.0,
                                               GridBagConstraints.CENTER,
                                               GridBagConstraints.NONE,
                                               new Insets(5, 5, 5, 20), 0,
                                               0));
        panelResource.setLayout(layoutResource);
        panelResource.add(textResource,
                          new GridBagConstraints(0, 0, 1, 1, 0.1, 0.1,
                                                 GridBagConstraints.WEST,
                                                 GridBagConstraints.HORIZONTAL,
                                                 new Insets(5, 5, 5, 5), 0,
                                                 0));
        panelResource.add(buttonResource,
                          new GridBagConstraints(1, 0, 1, 2, 0.0, 0.0,
                                                 GridBagConstraints.CENTER,
                                                 GridBagConstraints.NONE,
                                                 new Insets(5, 5, 5, 20), 0,
                                                 0));
        this.setLayout(layoutMain);
        this.add(panelSource,
                 new GridBagConstraints(0, 0, 1, 1, 0.1, 0.1,
                                        GridBagConstraints.CENTER,
                                        GridBagConstraints.BOTH,
                                        new Insets(5, 5, 5, 5), 0, 0));
        this.add(panelResource,
                 new GridBagConstraints(0, 1, 1, 1, 0.1, 0.1,
                                        GridBagConstraints.CENTER,
                                        GridBagConstraints.BOTH,
                                        new Insets(5, 5, 5, 5), 0, 0));
    }

    /**
     * Method declaration
     *
     *
     * @see
     */
    private void pmInit() {
        Border border = BorderFactory.createEtchedBorder();

        borderSource = BorderFactory.createTitledBorder(border);
        borderSource.setTitle(res.getString("borderSource_Title"));
        borderResource = BorderFactory.createTitledBorder(border);
        borderResource.setTitle(res.getString("borderResource_Title"));
        panelSource.setBorder(borderSource);
        panelResource.setBorder(borderResource);
        buttonListener = new LocalButtonListener();
        buttonResource.addActionListener(buttonListener);
        buttonSource.addActionListener(buttonListener);
        textResource.setText(new String());
        textSource.setText(new String());
    }

    /**
     * Method declaration
     *
     *
     * @see
     */
    private void browseForFile(JTextField field, String title) {
        File destDir = null;

        destDir = SwingUtil.getDirectoryChoice(this, field.getText(), title);
        if (destDir != null) {
            field.setText(PathHandle.createPathString(destDir));
        }
    }

    /**
     * Class declaration
     *
     *
     * @author Paul Mahar
     */
    private class LocalButtonListener implements ActionListener {

        public void actionPerformed(ActionEvent event) {
            Object source = event.getSource();

            if (source == buttonSource) {
                browseForFile(textSource,
                              res.getString("Select_source_path"));
            } else {
                browseForFile(textResource,
                              res.getString("Select_resource_path"));
            }
        }

    }
}
