/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 * Paul Mahar
 *
 */
package org.enhydra.kelp.common.xmlc;

// Enhydra XMLC imports
import org.enhydra.xml.xmlc.XMLCException;

// ToolBox imports
import org.enhydra.tool.common.FileUtil;
import org.enhydra.tool.common.PathHandle;

// Kelp imports
import org.enhydra.kelp.common.AbstractNodeBuilder;
import org.enhydra.kelp.common.Constants;
import org.enhydra.kelp.common.Writer;
import org.enhydra.kelp.common.event.WriteListener;
import org.enhydra.kelp.common.map.Mapper;
import org.enhydra.kelp.common.node.OtterTextFileNode;
import org.enhydra.kelp.common.node.OtterImageFileNode;
import org.enhydra.kelp.common.node.OtterProject;
import org.enhydra.kelp.common.node.OtterNode;
import org.enhydra.kelp.common.node.OtterXMLCNode;


// Standard imports
import java.io.IOException;
import java.io.PrintWriter;
import java.io.File;
import java.util.ResourceBundle;
import java.text.MessageFormat;

public class XMLCNodeBuilder extends AbstractNodeBuilder {
    static ResourceBundle res =
        ResourceBundle.getBundle("org.enhydra.kelp.common.Res");    // nores

    /**
     * Create an instance with listeners ready to
     * recieve output from XMLC.
     *
     *
     * @param listeners
     * A listener implements an onWrite method that is called
     * whenever XMLC writes output. The writes are captured
     * for display in a wizard.
     */
    public XMLCNodeBuilder(WriteListener[] listeners) {
        super(listeners);
    }

    public XMLCNodeBuilder(WriteListener listener, boolean buffer) {
        super(listener, buffer);
    }


    /**
     * Loop through all selected HTML files and call XMLC for each
     * and write out exceptions to any registered listeners.
     */
    public void buildImpl() {
        if (getNodes().length >= 1) {
            makeSelected();
        }
    }

    //
    private void makeSelected() {
        int        allCount = getNodes().length;
        int        selectCount = 0;
        double     selectIndex = 0.0;
        XMLCRunner runner = new XMLCRunner();

        refreshProgress(5, res.getString("XMLC_starting_"));
        runner.setTraceWriter(getEchoWriter());
        for (int i = 0; i < allCount; i++) {
            OtterXMLCNode xmlcNode = null;

            xmlcNode = (OtterXMLCNode) getNodes()[i];
            if (xmlcNode.isSelected()) {
                selectCount++;
            }
        }
        for (int i = 0; i < allCount; i++) {
            OtterXMLCNode xmlcNode = null;

            xmlcNode = (OtterXMLCNode) getNodes()[i];
            xmlcNode.setException(null);
            if (!isFresh()) {
                break;
            }
            try {
                if (xmlcNode.isSelected() && isFresh()) {
                    File       sourceDoc = null;
                    PathHandle sourcePath = null;
                    String     message = new String();
                    int        progValue = 0;
                    Integer[]  counts = new Integer[0];

                    selectIndex++;
                    progValue = (int) (5.0
                                       + (80.0
                                          * (selectIndex / selectCount)));
                    counts = new Integer[2];
                    counts[0] = new Integer((int) selectIndex);
                    counts[1] = new Integer(selectCount);
                    message =
                        MessageFormat.format(res.getString("XMLC_processing"),
                                             counts);
                    refreshProgress(progValue, message);
                    sourcePath =
                        PathHandle.createPathHandle(xmlcNode.getMetaDataHandler().getInputDocument());
                    if (sourcePath.getFile().exists()) {
                        if (isEcho()) {
                            getEchoWriter().println(res.getString("Compiling_"));
                            getEchoWriter().println(Constants.TAB4
                                                    + sourcePath.getPath());
                        }
                        xmlcNode.preCompile();
                        runner.setNode(xmlcNode);
                        runner.compile();
                    } else {
                        if (isEcho()) {
                            getEchoWriter().println(res.getString("Source_document_not")
                                                    + sourcePath.getPath());
                        }
                    }
                }
            } catch (XMLCException e) {
                e.printStackTrace();
                xmlcNode.setException(e);
                if (isEcho()) {
                    getEchoWriter().println(new String());
                }
            } catch (Throwable e) {
                xmlcNode.setException(e);
                if (isEcho()) {
                    getEchoWriter().println(new String());
                }
            }
        }
        if (isFresh()) {
            refreshProgress(90, res.getString("XMLC_gathering"));
            if (isEcho()) {
                printResults();
            }
            printExceptions();
            refreshProgress(95, res.getString("XMLC_complete"));
            if (isEcho()) {
                getEchoWriter().println(new String());
                getEchoWriter().println(res.getString("_XMLC_complete_"));
                getEchoWriter().println(new String());
            }
        } else {
            refreshProgress(90, res.getString("XMLC_interrupted"));
            printExceptions();
            getEchoWriter().println(new String());
            getEchoWriter().println(res.getString("_XMLC_interrupted_"));
            getEchoWriter().println(new String());
        }
        refreshProgress(100, res.getString("XMLC_complete"));
    }

    /**
     * Determine what files were created and write out the
     * compilation results to any registered listeners.
     */
    public void printResults() {
        int        eCount = 0;
        int        totalCount = getNodes().length;
        int        selectCount = 0;
        PathHandle path = null;

        getEchoWriter().println(new String());
        getEchoWriter().println(res.getString("Gathering_results_"));
        for (int i = 0; i < totalCount; i++) {
            if (getNodes()[i].isSelected()
                    && getNodes()[i].getException() != null) {
                eCount++;
            }
        }
        sleepInSeconds(0.5);
        for (int i = 0; i < totalCount; i++) {
            if (getNodes()[i].isSelected()) {
                selectCount++;
            }
        }
        if (eCount < selectCount) {
            getEchoWriter().println(new String());
            getEchoWriter().println(res.getString("New_java_files_"));
            for (int i = 0; i < totalCount; i++) {
                OtterXMLCNode xmlcNode = (OtterXMLCNode) getNodes()[i];

                sleepInSeconds(0.1);
                if (!isFresh()) {
                    break;
                }
                if (xmlcNode.isSelected() && xmlcNode.isNewJavaFound()) {
                    xmlcNode.replaceGeneratedSource(xmlcNode.getMetaDataHandler().getRecompilation());
                    path =
                        PathHandle.createPathHandle(xmlcNode.getMetaDataHandler().getJavaClassSource());
                    getEchoWriter().println(Constants.TAB2 + path.getPath());
                }
            }
        }
        sleepInSeconds(0.1);
    }

}
