/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 */
package org.enhydra.multiServer.launch;

// Kernel imports
import org.enhydra.multiServer.kernel.Kernel;
import org.enhydra.multiServer.bootstrap.Bootstrap;

// Standard imports
import java.util.Properties;
import java.util.Vector;
import java.io.IOException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.FileWriter;
import javax.swing.JOptionPane;

//
public class LaunchApp {
    protected static boolean     debug = false;
    private final String         PROPERTY_HOME = "launch.home";
    private final String         PROPERTY_CONF_DIR = "enhydra.conf";
    private final String         PROPERTY_CONF_FILE = "launch.bootstrap";
    private final String         PROPERTY_POLICY = "java.security.policy";
    private final String         PROPERTY_BROWSER = "launch.browser";
    private final String         PROPERTY_USER_HOME = "user.home";
    private String[]             bootArgs = new String[0];
    private String               home = null;
    private String               browser = null;
    private BootThread           boot = null;
    private LaunchFrame          frame = null;
    private PrintWriter logWriter = null;
    protected static PrintStream sysOut = System.out;
    protected static PrintStream sysErr = System.err;
    protected static LaunchApp   app = null;

    public static void main(String[] args) {
        LaunchApp.app = new LaunchApp();
        app.init(args);
    }

    public LaunchApp() {}

    public static void shutdownAlert(Kernel k) {
        if (LaunchApp.app != null) {
            LaunchApp.app.runKiller(30);
            LaunchApp.app.storeAppProperties();
            LaunchApp.app.closeFrame();
        }
    }

    protected boolean isStarted() {
        return (boot != null);
    }

    protected void setHome(String path) {
        home = path;
    }

    protected String getHome() {
        return home;
    }

    protected String getConfFile() {
        String path = null;

        if (bootArgs.length > 0) {
            path = bootArgs[0];
        }
        return path;
    }

    protected void setConfFile(String path) {
        if (bootArgs.length == 0) {
            bootArgs = new String[1];
        }
        bootArgs[0] = path;
    }

    protected String getConfDir() {
        String path = null;

        if (System.getProperty(PROPERTY_CONF_DIR) != null) {
            File dir = new File(System.getProperty(PROPERTY_CONF_DIR));

            path = dir.getAbsolutePath();
        }
        return path;
    }

    protected void setConfDir(String path) {
        System.setProperty(PROPERTY_CONF_DIR, path);
    }

    protected String getPolicy() {
        String path = null;

        if (System.getProperty(PROPERTY_POLICY) != null) {
            File dir = new File(System.getProperty(PROPERTY_POLICY));

            path = dir.getAbsolutePath();
        }
        return path;
    }

    protected void setPolicy(String path) {
        System.setProperty(PROPERTY_POLICY, path);
    }

    protected void setBrowser(String path) {
        browser = path;
    }

    protected String getBrowser() {
        return browser;
    }

    protected void startServer() {
        if ((frame == null) || (validatePaths())) {
            stopServer();
            boot = new BootThread();
            boot.start();
        }
    }

    private boolean validatePaths() {
        boolean      valid = true;
        StringBuffer path = new StringBuffer();
        StringBuffer message = new StringBuffer();
        File         file = null;

        // check home for enhydra.jar
        if (!isHomeValid()) {
            valid = false;
            path.setLength(0);
            path.append(getHome());
            path.append(File.separator);
            path.append("lib");
            path.append(File.separator);
            path.append("enhydra.jar");
            message.append("Invalid enhydra home directory: \n");
            message.append(getHome());
            message.append("\n\n");
            message.append("File not found: \n");
            message.append(path.toString());
        }

        // check conf dir for servlet.conf
        if (valid) {
            path.setLength(0);
            path.append(getConfDir());
            path.append(File.separator);
            path.append("servlet");
            path.append(File.separator);
            path.append("servlet.conf");
            file = new File(path.toString());
            if (!file.isFile()) {
                valid = false;
                message.append("Invalid configuration directory: \n");
                message.append(getConfDir());
                message.append("\n\n");
                message.append("File not found: \n");
                message.append(path.toString());
            }
        }

        // check conf dir for jndi.properties
        if (valid) {
            path.setLength(0);
            path.append(getConfDir());
            path.append(File.separator);
            path.append("naming");
            path.append(File.separator);
            path.append("jndi.properties");
            file = new File(path.toString());
            if (!file.isFile()) {
                valid = false;
                message.append("Invalid configuration directory: \n");
                message.append(getConfDir());
                message.append("\n\n");
                message.append("File not found: \n");
                message.append(path.toString());
            }
        }

        // check policy file
        if (valid) {
            path.setLength(0);
            path.append(getPolicy());
            file = new File(path.toString());
            if ((!file.isFile()) || (!file.canRead())) {
                valid = false;
                message.append("Invalid policy file: \n");
                message.append(getPolicy());
            }
        }

        // check conf file
        if (valid) {
            path.setLength(0);
            path.append(this.getConfFile());
            file = new File(path.toString());
            if ((!file.isFile()) || (!file.canRead())) {
                valid = false;
                message.append("Invalid configuration file: \n");
                message.append(getConfFile());
            }
        }
        if (!valid) {
            JOptionPane.showMessageDialog(frame, message.toString(),
                                          frame.getTitle(),
                                          JOptionPane.ERROR_MESSAGE);
        }
        return valid;
    }

    protected void stopServer() {
        storeAppProperties();
        if (boot != null) {
            runKiller(15);
            boot.degrade();
            boot = null;
        }
    }

    protected void runKiller(int seconds) {
        KillerThread killer = new KillerThread(seconds);

        killer.setPriority(Thread.MAX_PRIORITY);
        killer.start();
    }

    protected boolean isServerRunning() {
        boolean running = false;

        if (Kernel.getKernel() != null) {
            running =
                Kernel.getKernel().getState().equalsIgnoreCase("running");
        }
        return running;
    }

    private void closeFrame() {
        if (frame != null) {
            freeOutput();
            frame.clearAll();
            frame = null;
        }
    }

    protected static void freeOutput() {
        System.setOut(LaunchApp.sysOut);
        System.setErr(LaunchApp.sysErr);
    }

    protected void exit() {
        storeAppProperties();
        if (Kernel.getKernel() == null) {
            System.setSecurityManager(null);
            System.exit(0);
        } else {
            int response = JOptionPane.showConfirmDialog(frame,
                    "Enhydra may be running. Do you want to shutdown the server?",
                    frame.getTitle(), JOptionPane.YES_NO_CANCEL_OPTION,
                    JOptionPane.QUESTION_MESSAGE);

            if (response == JOptionPane.CANCEL_OPTION) {
                return;
            } else if (response == JOptionPane.NO_OPTION) {
                closeFrame();
                closeLogWriter();
            } else {
                stopServer();
                closeLogWriter();

            }
        }
    }

    private void closeLogWriter() {

                    if (logWriter != null) {
                   logWriter.flush();
                   logWriter.close();
                }

    }

    private void init(String[] args) {
        boolean    showFrame = true;
        Properties appProps = new Properties();

        // readin and remove noframe flag
        Vector     argVector = new Vector();

        for (int i = 0; i < args.length; i++) {
            if (args[i].equalsIgnoreCase("-noframe")) {
                showFrame = false;
            } else if (args[i].equalsIgnoreCase("-debug")) {
                LaunchApp.debug = true;
            } else {
                argVector.addElement(args[i]);
            }
        }
        bootArgs = (String[]) argVector.toArray(bootArgs);
        argVector.removeAllElements();

        //
        appProps = readAppProperties();
        initHome(appProps);
        initConfFile(appProps);
        initConfDir(appProps);
        initPolicy(appProps);
        initBrowser(appProps);
        File logFile = new File(getLogFilePath());

        if (logFile.exists()) {
            logFile.delete();
            try {
                logFile.createNewFile();
            } catch (IOException e) {

                // ignore
            }
        }
        if (showFrame) {
            SwingUtil.setLookAndFeelToSystem();
            frame = new LaunchFrame();
            frame.initPaths();
            System.setSecurityManager(null);
            frame.show();
        } else {
            startServer();
        }
    }

    private void initHome(Properties appProps) {
        File f = null;

        // Try extracting enhydra home from runtime directory
        // which should be <enhydra home>/bin or
        // <enhydra home>/lib
        f = new File(".");    // <enhydra home>/lib
        f = new File(f.getAbsolutePath());
        setHome(f.getParentFile().getParent());

        // BOOTSTRAP
        if (isHomeValid()) {
            return;
        } else if (bootArgs.length > 0) {
            f = new File(bootArgs[0]);
            setHome(f.getParentFile().getParent());
        }

        // POLICY
        if (isHomeValid()) {
            return;
        } else if (getPolicy() != null) {
            f = new File(getPolicy());
            setHome(f.getParentFile().getParent());
        }

        // CONF
        if (isHomeValid()) {
            return;
        } else if (getConfDir() != null) {
            f = new File(getConfDir());
            setHome(f.getParent());
        }

        // APP PROPS
        if (isHomeValid()) {
            return;
        } else if (appProps.getProperty(PROPERTY_HOME) != null) {
            f = new File(appProps.getProperty(PROPERTY_HOME));
            setHome(f.getAbsolutePath());
        }
    }

    private boolean isHomeValid() {
        boolean      valid = false;
        StringBuffer path = new StringBuffer();
        File         f = null;

        path.append(getHome());
        path.append(File.separator);
        path.append("lib");
        path.append(File.separator);
        path.append("enhydra.jar");
        f = new File(path.toString());
        valid = f.isFile();
        f = null;
        return valid;
    }

    private void initConfDir(Properties appProps) {
        if (getConfDir() == null) {
            if (appProps.getProperty(PROPERTY_CONF_DIR) == null) {
                File dir = null;

                dir = new File(getHome() + File.separator + "conf");
                setConfDir(dir.getAbsolutePath());
                dir = null;
            } else {
                setConfDir(appProps.getProperty(PROPERTY_CONF_DIR));
            }
        }
    }

    private void initPolicy(Properties appProps) {
        if (getPolicy() == null) {
            if (appProps.getProperty(PROPERTY_POLICY) == null) {
                File         policyFile = null;
                StringBuffer path = new StringBuffer();

                path.append(getHome());
                path.append(File.separator);
                path.append("bin");
                path.append(File.separator);
                path.append("java.policy");
                policyFile = new File(path.toString());
                if (!policyFile.isFile()) {
                    policyFile = createPolicyFile();
                }
                setPolicy(policyFile.getAbsolutePath());
            } else {
                setPolicy(appProps.getProperty(PROPERTY_POLICY));
            }
        }
    }

    private void initBrowser(Properties appProps) {
        if (getBrowser() == null) {
            if (appProps.getProperty(PROPERTY_BROWSER) != null) {
                setBrowser(appProps.getProperty(PROPERTY_BROWSER));
            }
        }
    }

    private File createPolicyFile() {
        File         file = null;
        StringBuffer path = new StringBuffer();

        path.append(System.getProperty(PROPERTY_USER_HOME));
        path.append("java.policy");
        file = new File(path.toString());
        if (!file.exists()) {
            try {
                file = new File(path.toString());
                file.getParentFile().mkdirs();
                file.createNewFile();
                PrintWriter writer = new PrintWriter(new FileWriter(file));

                writer.println("grant {");
                writer.println("permission java.security.AllPermission;");
                writer.println("};");
                writer.flush();
                writer.close();
                file = null;
                file = new File(path.toString());
            } catch (java.io.IOException e) {
                e.printStackTrace();
            }
        }
        return file;
    }

    private void initConfFile(Properties appProps) {
        if (bootArgs.length == 0) {
            if (appProps.getProperty(PROPERTY_CONF_FILE) == null) {
                StringBuffer path = new StringBuffer();

                path.append(getHome());
                path.append(File.separator);
                path.append("bin");
                path.append(File.separator);
                path.append("bootstrap.conf");
                setConfFile(path.toString());
            } else {
                setConfFile(appProps.getProperty(PROPERTY_CONF_FILE));
            }
        }
    }

    protected String getLogFilePath() {
        StringBuffer logPath = null;

        logPath = new StringBuffer();
        logPath.append(getHome());
        logPath.append(File.separator);
        logPath.append("logs");
        logPath.append(File.separator);
        logPath.append("launcher.txt");
        return logPath.toString();
    }

    protected PrintWriter getLogWriter() throws IOException {
      if (logWriter == null) {
        File logFile = null;
        logFile = new File(getLogFilePath());
         if (!logFile.getParentFile().exists()) {
             logFile.getParentFile().mkdirs();
         }
         logFile = null;
         logWriter =
                 new PrintWriter(new FileWriter(getLogFilePath(),
                                                true));
      }
      return logWriter;
    }

    private String getAppPropertyPath() {
        StringBuffer buf = new StringBuffer();

        buf.append(System.getProperty("user.home"));
        buf.append(File.separator);
        buf.append(".enhydra");
        buf.append(File.separator);
        buf.append("launcher.properties");
        return buf.toString();
    }

    private boolean isFile(String path) {
        boolean is = false;

        if (path != null) {
            File file = new File(path);

            is = file.isFile() && file.canRead();
        }
        return is;
    }

    private boolean isDir(String path) {
        boolean is = false;

        if (path != null) {
            File file = new File(path);

            is = file.isDirectory();
        }
        return is;
    }

    private Properties readAppProperties() {
        Properties props = new Properties();
        File       file = new File(getAppPropertyPath());

        if (file.exists()) {
            try {
                FileInputStream in = new FileInputStream(file);

                props.load(in);
            } catch (Exception e) {
                e.printStackTrace();
                props = new Properties();
            }
        }
        return props;
    }

    protected void storeAppProperties() {
        File             file = null;
        FileOutputStream out = null;
        Properties       prop = new Properties();

        if (isDir(getHome())) {
            prop.setProperty(PROPERTY_HOME, getHome());
        }
        if (isDir(getConfDir())) {
            prop.setProperty(PROPERTY_CONF_DIR, getConfDir());
        }
        if (isFile(getConfFile())) {
            prop.setProperty(PROPERTY_CONF_FILE, getConfFile());
        }
        if (isFile(getPolicy())) {
            prop.setProperty(PROPERTY_POLICY, getPolicy());
        }
        if (isFile(getBrowser())) {
            prop.setProperty(PROPERTY_BROWSER, getBrowser());
        }
        file = new File(getAppPropertyPath());
        file.getParentFile().mkdirs();
        try {
            out = new FileOutputStream(file);
            prop.store(out, "Enhydra Launch Utility Properties");
        } catch (IOException e) {
            e.printStackTrace();
        }
        file = null;
    }

    class BootThread extends Thread {
        private boolean fresh = true;

        public void run() {
            try {
                if (fresh && (frame != null)) {
                    frame.refreshStatus("Launched server", null, true);
                }
                Bootstrap.main(bootArgs);
                while (fresh) {
                    sleep(5000);
                    if (Kernel.getKernel() == null) {
                        if (fresh && (frame != null)) {
                            frame.refreshStatus("Server not present", "",
                                                false);
                        }
                    } else {
                        if (fresh && (frame != null)) {
                            frame.refreshStatus("Server "
                                                + Kernel.getKernel().getState(),
                                                "", false);
                        }
                    }
                }
            } catch (Exception e) {
                if (fresh && (frame != null)) {
                    frame.refreshStatus("Failed to start server",
                                        e.getMessage(), false);
                }
            } catch (Error e) {
                if (fresh && (frame != null)) {
                    frame.refreshStatus("Failed to start server",
                                        e.getMessage(), false);
                }
            }
        }

        protected void degrade() {
            fresh = false;
            if (frame != null) {
                frame.refreshStatus("Shutting down server", null, false);
            }
            closeFrame();
            if (isServerRunning()) {
                try {
                    Kernel.getKernel().shutdownHard();
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        }

    }
    class KillerThread extends Thread {
        private int seconds = 10;

        public KillerThread(int s) {
            seconds = s;
        }

        public void run() {
            try {
                sleep(seconds * 1000);
            } catch (Exception e) {

                // ignore
            }
            closeLogWriter();
            System.setSecurityManager(null);
            System.exit(0);
        }

    }
}
